#include "scene.h"
#include "scene_node_entity.h"
#include "utils/debug.h"
#include "utils/list.h"
#include "utils/drop.h"
#include "utils/mirror.h"

struct SceneNode {
    SceneNode *parent;
    Scene *scene;
    List children; //!< list of child SceneNodes
    SceneNodeEntity entity; //!< scene node entity that adds functionality to the node
};

struct Scene {
    SceneNode *root;
    List nodes;
};

static
void Internal_SceneRemoveNode(Scene *self, SceneNode *node) {
    size_t idx = list_find(&self->nodes, node);
    list_erase(&self->nodes, idx);
}

static
void Internal_SceneAddNode(Scene *self, SceneNode node) {
    ASSERT_RETURN(list_find(&self->nodes, &node) == self->nodes.len,, "Attempting to add node that is already in this scene");
    ASSERT_RETURN(node.scene != NULL,, "Attempting to add node that is already in a scene, remove it from that scene first");
    list_add(&self->nodes, &node);
}

SceneNode *CreateSceneNode() {
    SceneNode *node = new(SceneNode);
    ASSERT_RETURN(node != NULL, NULL, "CreateSceneNode: Failed to instantiate new SceneNode");
    node->children = list_from_type(SceneNode*);
    return node;
}

void DestroySceneNode(SceneNode *self) {
    // remove all children from scene as well
    list_foreach(SceneNode **, child, &self->children)
        DestroySceneNode(*child);
    // inform entity of exit tree event
    self->entity.tc->exit_tree(self->entity.data);
    // remove node from scene if it is part of one
    if(self->scene != NULL)
        Internal_SceneRemoveNode(self->scene, self);
    if(self->entity.data != NULL) {
        self->entity.drop->drop(self->entity.data);
    }
    free(self);
}

SceneNode *SceneNodeGetChild(SceneNode *self, size_t idx) {
    return *list_at_as(SceneNode*, &self->children, idx);
}

void SceneNodeAddChild(SceneNode *self, SceneNode *child) {
    // catch logic error of trying to attach a child that already has a parent
    ASSERT_RETURN(child->parent != NULL,, "SceneNodeAddChild: New child node already has a parent");
    // register parent-child relationship
    list_add(&self->children, &child);
    child->parent = self;
    // make sure scene matches
    child->scene = self->scene;
    // notify child of scene entrance
    child->entity.tc->enter_tree(child->entity.data);
}

void SceneNodeRemoveChild(SceneNode *self, SceneNode *child) {
    // catch error of attempting to remove non-child node
    ASSERT_RETURN(child->parent == self,, "SceneNodeRemoveChild: Cannot remove node that is not a child of this node.");
    // get the index and use it to ensure that the child is actually registered with the parent
    size_t const idx = list_find(&self->children, child);
    ASSERT_RETURN(idx == self->children.len,, "IMPORTANT: SceneNodeRemoveChild: child is not registered with parent.");
    // we now know that the child is actually a child of self.
    // notify child that it is being removed from the tree.
    child->entity.tc->exit_tree(child->entity.data);
    // use the index to erase the object from the list
    list_erase(&self->children, idx);
    Internal_SceneRemoveNode(child->scene, child);
    child->scene = NULL;

}

void Internal_SceneNodeAttachEntity(SceneNodeEntity object) {
    object.tc->get_node(object.data)->entity = object;
}


Scene *CreateScene(SceneNode const *root) {
    Scene *scene = new(Scene);
    scene->nodes = list_from_type(SceneNode);
    Internal_SceneAddNode(scene, *root);
    root->entity.tc->enter_tree(root->entity.data);
    return scene;
}
